from django.contrib import admin
from django.db.models import QuerySet
from django.http import HttpRequest
from django.urls import reverse
from django.utils.safestring import mark_safe


class EditLinkToInlineObject(object):
    """
    Adapted from https://stackoverflow.com/a/22113967/5666845

    Usage:

    @admin.register(MyModel)
    class MyModelAdmin:
        pass

    class MyModelInline(EditLinkToInlineObject, admin.TabularInline):
        model = MyModel
        readonly_fields = ('edit_link', )

    @admin.register(MySecondModel)
    class MySecondModelAdmin(admin.ModelAdmin):
        inlines = (MyModelInline, )

    """

    def edit_link(self, instance):
        url = reverse(
            "admin:%s_%s_change"
            % (instance._meta.app_label, instance._meta.model_name),
            args=[instance.pk],
        )
        if instance.pk:
            return mark_safe(
                '<a href="{u}" target="_blank">edit</a>'.format(u=url)
            )
        else:
            return ""


class MeOrAllFilter(admin.SimpleListFilter):
    """
    Users can filter Models with a `user` field by the current user, or all users

    Usage:
        import MeOrAllFilter and add it to the ModelAdmin's list_filter

    Example:
    ```
    @admin.register(MyModel)
    class MyModelAdmin(admin.ModelAdmin):
        list_filter = [MeOrAllFilter,]
    ```

    """

    # based on https://docs.djangoproject.com/en/2.2/ref/contrib/admin/#django.contrib.admin.ModelAdmin.list_filter
    title = "User"
    parameter_name = "just_me"

    def lookups(self, request: HttpRequest, model_admin: admin.ModelAdmin):
        return [("true", "Just Me")]  # 'All' option automatically provided

    def queryset(self, request: HttpRequest, queryset: QuerySet):
        if self.value() == "true":
            queryset = queryset.filter(user=request.user)
        return queryset


# TODO (low priority): there's a more django-ic way to do this
def is_null_filter_factory(field: str):
    """
    Creates a filter for is/is not null for the specified field
    A proper FieldListFilter would do something a little more complicated
    usage:

    @admin.register(ActivityAssignment)
    class ActivityAssignmentAdmin(admin.ModelAdmin):
        list_display = [...]
        readonly_fields = [...]
        list_filter = [
            ('actual_end', admin.DateFieldListFilter),
            ('activity', admin.RelatedFieldListFilter),
            is_null_filter_factory('response'),
        ]

    """

    class IsNullFilter(admin.SimpleListFilter):
        # based on https://docs.djangoproject.com/en/2.2/ref/contrib/admin/#django.contrib.admin.ModelAdmin.list_filter
        title = f"{field} is null"
        parameter_name = f"{field}__isnull"

        def lookups(self, request: HttpRequest, model_admin: admin.ModelAdmin):
            return [("null", "Is Null"), ("not_null", "Is not Null")]

        def queryset(self, request: HttpRequest, queryset: QuerySet):
            if self.value() == "null":
                queryset = queryset.filter(**{field + "__isnull": True})
            elif self.value() == "not_null":
                queryset = queryset.exclude(**{field + "__isnull": True})
            return queryset

    return IsNullFilter


USER_SEARCH_FIELDS = (
    "user__username",
    "user__first_name",
    "user__last_name",
    "user__email",
)




from django.contrib.auth.models import User
from django.test.testcases import TestCase
from django.utils import timezone


from django_notification_system.models import (
    Notification, NotificationTarget, TargetUserRecord)
from django_notification_system.notification_handlers.email import (
    send_notification)


class TestCreateEmailNotification(TestCase):
    def setUp(self):
        self.user_with_target = User.objects.create_user(
            username='sadboi@gmail.com',
            first_name='Sad',
            last_name='Boi',
            password='Ok.',
            email='sadboi@gmail.com')

        self.target_user_record = TargetUserRecord.objects.create(
            user=self.user_with_target,
            target=NotificationTarget.objects.get(name='Email'),
            target_user_id='sadboi@gmail.com',
            description='Sad Bois Email',
            active=True
        )

        self.notification = Notification.objects.create(
            target_user_record=self.target_user_record,
            title="Hi.",
            body="<b>It me. Is it me?</b>",
            status='SCHEDULED',
            scheduled_delivery=timezone.now())

    def test_send_email(self):
        """
        Test proper functionality of send_email function.
        """
        pre_function_notification = Notification.objects.all()
        self.assertEqual(pre_function_notification[0].status,
                         'SCHEDULED')

        response_message = send_notification(self.notification)

        self.assertEqual(response_message, 'Email Successfully Sent')



# Generated by Django 3.1.3 on 2020-12-03 21:28

from django.db import migrations


def create_initial_targets(apps, schema_editor):
    """
    Initialize the database with the Targets that we have functions available for.
    """
    NotificationTarget = apps.get_model('django_notification_system', 'NotificationTarget')
    NotificationTarget.objects.get_or_create(
        name="Twilio",
        notification_module_name="twilio")
    NotificationTarget.objects.get_or_create(
        name="Email",
        notification_module_name="email")
    NotificationTarget.objects.get_or_create(
        name="Expo",
        notification_module_name="expo")


class Migration(migrations.Migration):

    dependencies = [
        ('django_notification_system', '0001_initial'),
    ]

    operations = [
        migrations.RunPython(create_initial_targets),
    ]



"""A Django Notification System Handler."""
from exponent_server_sdk import (
    PushClient,
    PushMessage,
    PushResponseError,
    PushServerError,
    DeviceNotRegisteredError,
)
from requests import HTTPError

from django.utils import timezone

from ..utils import check_and_update_retry_attempts


def send_notification(notification) -> str:
    """
    Send push notifications (Expo) to the target device using the Expo server.

    Args:
        notification (Notification): The Expo push notification to be sent.

    Returns:
        String: Whether the push notification has successfully sent, or an error message.
    """
    extra = prepare_extra(notification.extra)

    try:
        response = PushClient().publish(
            PushMessage(
                to=str(notification.target_user_record.target_user_id),
                title=notification.title,
                body=notification.body,
                data=extra["data"],
                sound=extra["sound"],
                ttl=extra["ttl"],
                expiration=extra["expiration"],
                priority=extra["priority"],
                badge=extra["badge"],
                channel_id=extra["channel_id"],
            )
        )
    except (PushServerError, HTTPError, ValueError) as e:
        check_and_update_retry_attempts(notification)
        return "{}: {}".format(type(e), e)

    return handle_push_response(notification, response)


def prepare_extra(extra):
    """
    Take in a JSON object from the Notification model instance and prepare a dictionary with all
    options available for PushMessage(). Loop through expected push_options and if the option
    exists in 'extra' JSON then it gets added to the dict with its value or else it gets added
    to the dictionary with None as its value.

    Args:
        extra (dict): JSON from the Notification model instance field 'extra'

    Returns:
        dict: Dictionary consisting of PushMessage() options
    """
    final_extra = {}
    push_options = [
        "data",
        "sound",
        "ttl",
        "expiration",
        "priority",
        "badge",
        "channel_id",
    ]
    for item in push_options:
        try:
            final_extra[item] = extra[item]
        except (KeyError, TypeError):
            final_extra[item] = None
    return final_extra


def handle_push_response(notification, response):
    """
    This function handles the push response requested in send_expo()

    Args:
        notification (Notification): The Expo push notification to be sent.
        response (PushResponse): The Expo push response

    Returns:
        str: Whether the push has successfully sent, or an error message.
    """
    try:
        # We got a response back, but we don't know whether it's an error yet.
        # This call raises errors so we can handle them with normal exception
        # flows.
        response.validate_response()
    except DeviceNotRegisteredError as e:
        target_user_record = notification.target_user_record
        target_user_record.active = False
        target_user_record.save()
        return "{}: {}".format(type(e), e)
    except PushResponseError as e:
        check_and_update_retry_attempts(notification)
        return "{}: {}".format(type(e), e)
    else:
        notification.status = notification.DELIVERED
        notification.attempted_delivery = timezone.now()
        notification.save()
        return "Notification Successfully Pushed!"



"""
This package contains all model definitions for the Django Notification System.
As a convience, the models are made available directly on the package namespace.
"""

from .notification import Notification
from .opt_out import NotificationOptOut
from .target import NotificationTarget
from .target_user_record import TargetUserRecord

__all__ = [
    "NotificationOptOut",
    "NotificationTarget",
    "TargetUserRecord",
    "Notification",
]



# Generated by Django 3.1.3 on 2020-12-09 21:26

from django.conf import settings
from django.db import migrations, models
import django.db.models.deletion
import uuid


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
    ]

    operations = [
        migrations.CreateModel(
            name='NotificationTarget',
            fields=[
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('modified_date', models.DateTimeField(auto_now=True)),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('name', models.CharField(max_length=15, unique=True)),
                ('notification_module_name', models.CharField(max_length=50)),
            ],
            options={
                'verbose_name_plural': 'Notification Targets',
                'db_table': 'notification_system_target',
            },
        ),
        migrations.CreateModel(
            name='TargetUserRecord',
            fields=[
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('modified_date', models.DateTimeField(auto_now=True)),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('target_user_id', models.CharField(max_length=200)),
                ('description', models.CharField(max_length=200)),
                ('active', models.BooleanField(default=True)),
                ('target', models.ForeignKey(on_delete=django.db.models.deletion.PROTECT, to='django_notification_system.notificationtarget')),
                ('user', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='notification_target_user_records', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name_plural': 'User In Notification Targets',
                'db_table': 'notification_system_target_user_record',
            },
        ),
        migrations.CreateModel(
            name='NotificationOptOut',
            fields=[
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('modified_date', models.DateTimeField(auto_now=True)),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('active', models.BooleanField(default=False)),
                ('user', models.OneToOneField(on_delete=django.db.models.deletion.CASCADE, related_name='notification_opt_out', to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'verbose_name_plural': 'Notification Opt Outs',
                'db_table': 'notification_system_opt_out',
            },
        ),
        migrations.CreateModel(
            name='Notification',
            fields=[
                ('created_date', models.DateTimeField(auto_now_add=True)),
                ('modified_date', models.DateTimeField(auto_now=True)),
                ('id', models.UUIDField(default=uuid.uuid4, editable=False, primary_key=True, serialize=False)),
                ('title', models.CharField(max_length=100)),
                ('body', models.TextField()),
                ('extra', models.JSONField(blank=True, default=dict, null=True)),
                ('status', models.CharField(choices=[('DELIVERED', 'Delivered'), ('DELIVERY FAILURE', 'Delivery Failure'), ('INACTIVE DEVICE', 'Inactive Device'), ('OPTED OUT', 'Opted Out'), ('RETRY', 'Retry'), ('SCHEDULED', 'Scheduled')], max_length=16)),
                ('scheduled_delivery', models.DateTimeField()),
                ('attempted_delivery', models.DateTimeField(blank=True, null=True)),
                ('retry_time_interval', models.PositiveIntegerField(default=0)),
                ('retry_attempts', models.PositiveIntegerField(default=0)),
                ('max_retries', models.PositiveIntegerField(default=3)),
                ('target_user_record', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, related_name='notifications', to='django_notification_system.targetuserrecord')),
            ],
            options={
                'verbose_name_plural': 'Notifications',
                'db_table': 'notification_system_notification',
            },
        ),
        migrations.AddConstraint(
            model_name='targetuserrecord',
            constraint=models.UniqueConstraint(fields=('user', 'target', 'target_user_id'), name='user target ids cannot be repeated'),
        ),
        migrations.AlterUniqueTogether(
            name='notification',
            unique_together={('target_user_record', 'scheduled_delivery', 'title', 'extra')},
        ),
    ]



import html2text
import socket
from smtplib import SMTPException

import django.core.mail
from django.conf import settings
from django.utils import timezone

from ..utils import check_and_update_retry_attempts


def send_notification(notification):
    """
    Send email notifications to the target device using the email server.

    Args:
        notification (Notification): The email notification to be sent.

    Returns:
        str: Whether the email has successfully sent, or an error message.
    """
    try:
        django.core.mail.send_mail(
            subject=notification.title,
            message=html2text.html2text(notification.body),
            html_message=notification.body,
            from_email=settings.NOTIFICATION_SYSTEM_TARGETS['email']['from_email'],
            recipient_list=[notification.target_user_record.target_user_id],
            fail_silently=False,
        )

    except SMTPException as e:
        # Update the notification to retry tomorrow if we are not at the
        # max amount of retries. SMTPEXceptions are usually the result of
        # hitting our daily limit of emails.
        check_and_update_retry_attempts(notification)
        return "Email could not be sent: {}".format(e)

    except socket.error as se:
        # Update the notification to retry in 90 minutes if we are not at the
        # max amount of retries. Socket errors are rare, sporadic and
        # inconsistent but usually resolved relatively quickly
        check_and_update_retry_attempts(notification, 90)
        return "Email could not be sent: {}".format(se)

    # If everything is fine, we update the notification
    # to DELIVERED
    notification.status = notification.DELIVERED
    notification.attempted_delivery = timezone.now()
    notification.save()
    return "Email Successfully Sent"



from twilio.rest import Client

from django.conf import settings
from django.utils import timezone

from ..utils import check_and_update_retry_attempts


def send_notification(notification):
    """
    Send Twilio notifications to the target device using the Twilio Client.

    Args:
        notification (Notification): The email notification to be sent.

    Returns:
        String: Whether the SMS has successfully sent, or an error message.
    """
    try:
        twilio_settings = settings.NOTIFICATION_SYSTEM_TARGETS['twilio_sms']
        twilio_account_sid = twilio_settings['account_sid']
        twilio_auth_token = twilio_settings['auth_token']

        twilio_sender = twilio_settings['sender']
        twilio_receiver = notification.target_user_record.target_user_id

        client = Client(twilio_account_sid, twilio_auth_token)

        client.messages.create(
            body=notification.body,
            from_=twilio_sender,
            to=twilio_receiver)

    except Exception as e:
        check_and_update_retry_attempts(notification)
        return ("{}: {}".format(type(e), e))

    else:
        notification.status = notification.DELIVERED
        notification.attempted_delivery = timezone.now()
        notification.save()
        return('SMS Successfully sent!')



import importlib
import inspect
import os
from os import path

from django.conf import settings
from django.core.management.base import BaseCommand
from django.db.models import Q
from django.utils import timezone

from ...models import Notification
from ...notification_handlers.email import send_notification as send_email
from ...notification_handlers.twilio import send_notification as send_twilio
from ...notification_handlers.expo import send_notification as send_expo


class Command(BaseCommand):
    """
    Push all SCHEDULED notifications with a scheduled_delivery before the current date_time
    """

    help = __doc__

    __function_table = {
        "expo": send_expo,
        "twilio": send_twilio,
        "email": send_email
    }

    @classmethod
    def _load_function_table(cls):
        """
        This function will get our function table populated with all available `send_notification`
        functions.
        """
        if hasattr(settings, "NOTIFICATION_SYSTEM_HANDLERS"):
            for directory in settings.NOTIFICATION_SYSTEM_HANDLERS:
                try:
                    for file in os.listdir(path.join(directory)):
                        if "__init__" in file:
                            # Ignore the init file
                            continue
                        try:
                            # Create the module spec
                            module_spec = importlib.util.spec_from_file_location(
                                file, f"{directory}/{file}")
                            # Create a new module  based  on the spec
                            module = importlib.util.module_from_spec(module_spec)
                            # An abstract method that executes the module
                            module_spec.loader.exec_module(module)
                            #  Get the actual function
                            real_func = getattr(module, "send_notification")
                            # Add it to our dictionary of functions
                            notification_system = file.partition(".py")[0]
                            cls.__function_table[notification_system] = real_func
                        except (ModuleNotFoundError, AttributeError):
                            pass
                except FileNotFoundError:
                    # the directory provided in the settings file does not exist
                    pass

    def handle(self, *args, **options):
        # Load the function table
        self._load_function_table()

        # Get all SCHEDULED and RETRY notifications with a
        # scheduled_delivery before the current date_time
        notifications = Notification.objects.filter(
            Q(status="SCHEDULED") | Q(status="RETRY"),
            scheduled_delivery__lte=timezone.now(),
        )

        # excludes all notifications where the user has NotificationOptOut object with has_opted_out=True
        notifications = notifications.exclude(target_user_record__user__notification_opt_out__active=True)

        # Loop through each notification and attempt to push it
        for notification in notifications:
            print(
                f"{notification.target_user_record.user.username} - {notification.scheduled_delivery} - {notification.status}")
            print(f"{notification.title} - {notification.body}")

            if not notification.target_user_record.active:
                notification.status = Notification.INACTIVE_DEVICE
                notification.save()
            else:
                notification_type = (
                    notification.target_user_record.target.notification_module_name
                )
                try:
                    # Use our function table to call the appropriate sending function
                    response_message = self.__function_table[notification_type](notification)
                except KeyError:
                    print(
                        f"invalid notification target name {notification.target_user_record.target.name}")
                else:
                    # The notification was sent successfully
                    print(response_message)
                    print("*********************************")



from django.contrib.auth.models import User
from django.test.testcases import TestCase
from django.utils import timezone

from django_notification_system.models import (
    Notification, NotificationTarget, TargetUserRecord, NotificationOptOut)
from django_notification_system.notification_creators.twilio import (
    create_notification)
from django_notification_system.exceptions import UserIsOptedOut, UserHasNoTargetRecords


class TestCreateNotification(TestCase):
    def setUp(self):
        self.user_with_targets = User.objects.create_user(
            username='sadboi@gmail.com',
            email='sadboi@gmail.com',
            first_name='Sad',
            last_name='Boi',
            password='Ok.')

        self.user_without_target = User.objects.create_user(
            username='skeeter@gmail.com',
            email='skeeter@gmail.com',
            first_name='Skeeter',
            last_name='Skeetington',
            password='Sure.')

        self.target, created = NotificationTarget.objects.get_or_create(
            name='Twilio',
            notification_module_name='twilio')

        self.user_target1 = TargetUserRecord.objects.create(
            user=self.user_with_targets,
            target=self.target,
            target_user_id='2917471274',
            description='Happy Phone')

        self.user_target2 = TargetUserRecord.objects.create(
            user=self.user_with_targets,
            target=self.target,
            target_user_id='2917471271',
            description='Happier Phone')

    def test_successfully_create_expo_notificationss(self):
        """
        This test checks that notifications are created for all
        user targets associated with the user.
        """
        pre_function_notifications = Notification.objects.all()
        self.assertEqual(len(pre_function_notifications), 0)

        create_notification(user=self.user_with_targets,
                            title="Wow",
                            body="You really did it!")

        post_function_notifications = Notification.objects.all()
        self.assertEqual(len(post_function_notifications), 2)

    def test_successfully_create_expo_notif_w_opt_out(self):
        """
        This test checks that notifications are created for all
        user targets associated with the user.
        """
        pre_function_notifications = Notification.objects.all()
        self.assertEqual(len(pre_function_notifications), 0)

        NotificationOptOut.objects.create(user=self.user_with_targets)

        create_notification(user=self.user_with_targets,
                            title="Wow",
                            body="You really did it!")

        post_function_notifications = Notification.objects.all()
        self.assertEqual(len(post_function_notifications), 2)

    def test_no_notifications_created__no_target(self):
        """
        If a user has no targets then no notifications can
        be created.
        """
        pre_function_notifications = Notification.objects.all()
        self.assertEqual(len(pre_function_notifications), 0)

        try:
            create_notification(user=self.user_without_target,
                                title="Wow",
                                body="You really did it!")
        except UserHasNoTargetRecords:
            pass

        post_function_notifications = Notification.objects.all()
        self.assertEqual(len(post_function_notifications), 0)

    def test_no_notifications_created__opt_out(self):
        """
        This test checks that notifications are created for all
        user targets associated with the user.
        """
        pre_function_notifications = Notification.objects.all()
        self.assertEqual(len(pre_function_notifications), 0)

        NotificationOptOut.objects.create(user=self.user_with_targets, active=True)

        try:
            create_notification(user=self.user_with_targets,
                                title="Wow",
                                body="You really did it!")
        except UserIsOptedOut:
            pass

        post_function_notifications = Notification.objects.all()
        self.assertEqual(len(post_function_notifications), 0)


from datetime import datetime
from django_notification_system.utils import (
    check_for_user_opt_out,
    user_notification_targets,
)

from django.contrib.auth.models import User
from django.template.loader import get_template
from django.utils import timezone

from ..models import Notification
from ..exceptions import (
    NotificationsNotCreated,
    UserHasNoTargetRecords,
    UserIsOptedOut,
)


def create_notification(
    user: User,
    title: str,
    body: str = "",
    scheduled_delivery: datetime = None,
    retry_time_interval: int = 1440,
    max_retries: int = 3,
    quiet=False,
    extra: dict = None,
) -> None:
    """
    This function will generate an email notification.

    Args:
        user (User): The user to whom the notification will be sent.
        title (str): The title for the notification.
        body (str, optional): Body of the email. Defaults to a blank string if not given.
            Additionally, if this parameter is not specific AND "template_name" is present
            in `extra`, an attempt will be made to generate the body from that template.
        scheduled_delivery (datetime, optional): When to delivery the notification. Defaults to immediately.
        retry_time_interval (int, optional): When to retry sending the notification if a delivery failure occurs. Defaults to 1440 seconds.
        max_retries (int, optional): Maximum number of retry attempts. Defaults to 3.
        quiet (bool, optional): Suppress exceptions from being raised. Defaults to False.
        extra (dict, optional): User specified additional data that will be used to
            populate an HTML template if "template_name" is present inside.

    Raises:
        UserIsOptedOut: When the user has an active opt-out.
        UserHasNoTargetRecords: When the user has no eligible targets for this notification type.
        NotificationsNotCreated: When the notifications could not be created.
    """

    try:
        check_for_user_opt_out(user=user)
    except UserIsOptedOut:
        if quiet:
            return
        else:
            raise UserIsOptedOut()

    target_user_records = user_notification_targets(user=user, target_name="Email")

    if not target_user_records:
        if quiet:
            return
        else:
            raise UserHasNoTargetRecords()

    if scheduled_delivery is None:
        scheduled_delivery = timezone.now()

    # Determine the body of the email. Preference is
    if body:
        email_body = body
    elif "template_name" in extra:
        # TODO: Look into how this function works and if we can just instruct people to include email templates in the TEMPLATE_DIRS setting.
        template = get_template(extra["template_name"])
        email_body = template.render(extra)
    else:
        raise ValueError(
            "You must either specify a `body` value or include 'template_name' in `extra` to create an email notification."
        )

    notifications_created = []
    for target_user_record in target_user_records:
        notification, created = Notification.objects.get_or_create(
            target_user_record=target_user_record,
            title=title,
            scheduled_delivery=scheduled_delivery,
            defaults={
                "body": email_body,
                "status": "SCHEDULED",
                "retry_time_interval": retry_time_interval,
                "max_retries": max_retries,
            },
        )

        if created:
            notifications_created.append(notification)

    if not notifications_created:
        if quiet:
            return
        else:
            raise NotificationsNotCreated()



import uuid

from django.conf import settings
from django.db import models
from django.db import transaction

from .abstract import CreatedModifiedAbstractModel
from .target import NotificationTarget


class TargetUserRecord(CreatedModifiedAbstractModel):
    """
    Definition of a TargetUserRecord.

    Each user will have a unique ID for each notification target,
    which is how we identify the individual who will receive the
    notification.

    For example, for an email target, we need to store the
    user's email address.

    Attributes
    ----------
    id : UUID
        The unique UUID of the record.
    user : Django User / Custom User Instance
        The User/Custom User instance associated with this record.
    target: Foreign Key
        The associated target instance.
    target_user_id : str
        The ID used in the target to uniquely identify the user.
    description : str
        A human friendly note about the user target.
    active : boolean
        Indicator of whether user target is active or not. For example,
        we have an outdated email record for a user.
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    user = models.ForeignKey(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="notification_target_user_records",
    )
    target = models.ForeignKey(NotificationTarget, on_delete=models.PROTECT)
    target_user_id = models.CharField(max_length=200)
    description = models.CharField(max_length=200)
    active = models.BooleanField(default=True)

    class Meta:
        db_table = "notification_system_target_user_record"
        verbose_name_plural = "User In Notification Targets"
        constraints = [
            models.UniqueConstraint(
                fields=["user", "target", "target_user_id"],
                name="user target ids cannot be repeated",
            )
        ]

    def __str__(self):
        return "{}: {}".format(self.user.username, self.description)



import uuid

from django.db import models

from .abstract import CreatedModifiedAbstractModel


class NotificationTarget(CreatedModifiedAbstractModel):
    """
    Definition of a Notification Target.

    A target represents something that can receive a
    notication from our system.

    Attributes
    ----------
    id : UUID
        The unique UUID of the record.
    name : CharField
        The human friendly name for the target.
    notification_module_name : str
        The name of the module in the notification_creators directory which
        will be used to create notifications for this target.
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    name = models.CharField(max_length=15, unique=True)
    notification_module_name = models.CharField(max_length=50)

    def __str__(self):
        return self.name

    class Meta:
        db_table = "notification_system_target"
        verbose_name_plural = "Notification Targets"




from unittest.mock import patch
from smtplib import SMTPException

from django.contrib.auth.models import User
from django.test.testcases import TestCase
from django.utils import timezone


from django_notification_system.models import (
    Notification, NotificationTarget, TargetUserRecord)

from django_notification_system.notification_handlers.email import send_notification


class TestSMTPExceptionEmailNotification(TestCase):
    def setUp(self):
        self.user_with_target = User.objects.create_user(
            username='sadboi@gmail.com',
            first_name='Sad',
            last_name='Boi',
            password='Ok.',
            email='sadboi@gmail.com')

        self.target, created = NotificationTarget.objects.get_or_create(
            name='Email',
            notification_module_name='email'
        )

        self.user_target = TargetUserRecord.objects.create(
            user=self.user_with_target,
            target=self.target,
            target_user_id='sadboi@gmail.com',
            description="Sad Boi's Email",
            active=True
            )

        self.notification = Notification.objects.create(
            target_user_record=self.user_target,
            title="Hi.",
            body="<b>It me. Is it me?</b>",
            status='SCHEDULED',
            scheduled_delivery=timezone.now(),
            max_retries=2)

    def test_smtp_exception(self):
        """
        Test that an smtp_exception is correctly handled.
        """
        with patch('django.core.mail.send_mail') as fake_send:
            fake_send.side_effect = SMTPException("No server")

            # Assert exception is correctly handled
            response_message = send_notification(
                self.notification)

            self.assertEqual(response_message,
                             'Email could not be sent: No server')
            # Assert DELIVERY_FAILURE after max_retries hit
            send_notification(self.notification)
            self.assertEqual(self.notification.status, 'DELIVERY FAILURE')



from datetime import datetime
from django_notification_system.utils import (
    check_for_user_opt_out,
    user_notification_targets,
)

from django.contrib.auth.models import User
from django.utils import timezone

from ..models import Notification
from ..exceptions import (
    NotificationsNotCreated,
    UserHasNoTargetRecords,
    UserIsOptedOut,
)


def create_notification(
    user: User,
    title: str,
    body: str,
    scheduled_delivery: datetime = None,
    retry_time_interval: int = 1440,
    max_retries: int = 3,
    quiet=False,
    extra: dict = None,
) -> None:
    """
    This function will generate a Twilio SMS notification.

    Args:
        user (User): The user to whom the push notification will be sent.
        title (str): The title for the push notification.
        body (str, optional): The body of the sms notification.
        scheduled_delivery (datetime, optional): Defaults to immediately.
        retry_time_interval (int, optional): Delay between send attempts. Defaults to 60 seconds.
        max_retries (int, optional): Maximum number of retry attempts for delivery. Defaults to 3.
        quiet (bool, optional): Suppress exceptions from being raised. Defaults to False.
        extra (dict, optional): Defaults to None.

    Raises:
        UserIsOptedOut: When the user has an active opt-out.
        UserHasNoTargetRecords: When the user has no eligible targets for this notification type.
        NotificationsNotCreated: When the notifications could not be created.
    """
    try:
        check_for_user_opt_out(user=user)
    except UserIsOptedOut:
        if quiet:
            return
        else:
            raise UserIsOptedOut()

    target_user_records = user_notification_targets(user=user, target_name="Twilio")

    if not target_user_records:
        if quiet:
            return
        else:
            raise UserHasNoTargetRecords()

    if scheduled_delivery is None:
        scheduled_delivery = timezone.now()

    notifications_created = []
    for target_user_record in target_user_records:
        # JSONField casts None to {}, so we have to check if the server value = {}
        if extra is None:
            extra = {}  # I hate casting so much
        notification, created = Notification.objects.get_or_create(
            target_user_record=target_user_record,
            title=title,
            scheduled_delivery=scheduled_delivery,
            extra=extra,
            defaults={
                "body": body,
                "status": 'SCHEDULED',
                "retry_time_interval": retry_time_interval,
                "max_retries": max_retries
            }
        )

        if created:
            notifications_created.append(notification)

    if not notifications_created:
        if quiet:
            return
        else:
            raise NotificationsNotCreated()



"""Application Wide Abstract Model Definitions"""

from django.db import models


class CreatedModifiedAbstractModel(models.Model):
    """
    Abstract base model that is used to add `created_date`
    and `modified_date` fields to all descendant models.
    """

    created_date = models.DateTimeField(auto_now_add=True)
    modified_date = models.DateTimeField(auto_now=True)

    class Meta:
        abstract = True


from datetime import timedelta
from unittest.mock import patch

from django.contrib.auth.models import User
from django.core.management import call_command
from django.test import TestCase
from django.utils import timezone
from exponent_server_sdk import PushResponse
from six import StringIO

from django_notification_system.models import (
    NotificationTarget, TargetUserRecord, Notification, NotificationOptOut)
from django_notification_system.notification_handlers.expo import handle_push_response
from ...mock_exponent_server_sdk import MockPushClient


@patch('django_notification_system.notification_handlers.expo.PushClient', new=MockPushClient)
class TestCommand(TestCase):
    def setUp(self):
        self.user_with_notifications = User.objects.create_user(
            username="Eggless",
            email="eggless@gmail.com",
            first_name="Egg",
            last_name="Less",
            password="ImpressivePassword")

        self.dev_user = User.objects.create_user(
            username="Danglesauce",
            email="danglesauce@gmail.com",
            first_name="Dangle",
            last_name="Sauce",
            password="ImpressivePassword")

        self.user_target = TargetUserRecord.objects.create(
            user=self.user_with_notifications,
            target=NotificationTarget.objects.get(name='Expo'),
            target_user_id='ExponentPushToken[ByAAmjPd96SUb1Is5eUzXX]',
            description="It's Expo",
            active=True
        )

        self.dev_user_target = TargetUserRecord.objects.create(
            user=self.dev_user,
            target=NotificationTarget.objects.get(name="Expo"),
            target_user_id='ExponentPushToken[ByAAmjPd96SUb1Is5eUzXX]',
            description='Dangle Expo',
            active=True
        )

        self.user_target_email = TargetUserRecord.objects.create(
            user=self.user_with_notifications,
            target=NotificationTarget.objects.get(name='Email'),
            target_user_id=self.user_with_notifications.email,
            description='Its email',
            active=True,
        )

        self.user_target_twilio = TargetUserRecord.objects.create(
            user=self.user_with_notifications,
            target=NotificationTarget.objects.get(name='Twilio'),
            target_user_id='6766676677',
            description='Test Phone #',
            active=True
        )

        self.notification = Notification.objects.create(
            target_user_record=self.user_target,
            status=Notification.SCHEDULED,
            title="Title",
            body="Body of the message",
            extra={
                "sound": "default",
                "data": {"Junk": "Junk"},
                "priority": "high",
                "ttl": "3",
                "expiration": "1486080000",
                "badge": "3",
            },
            scheduled_delivery="2018-10-24T12:42:13-04:00",
        )

        self.dev_notification = Notification.objects.create(
            target_user_record=self.dev_user_target,
            status=Notification.SCHEDULED,
            title="Title",
            body="Whoops",
            scheduled_delivery=timezone.now() - timedelta(1),
        )

        self.notification_email = Notification.objects.create(
            target_user_record=self.user_target_email,
            status=Notification.SCHEDULED,
            title="Title",
            body="<p>Body of the message</p>",
            scheduled_delivery=timezone.now() - timedelta(1),
        )

        self.notification_twilio = Notification.objects.create(
            target_user_record=self.user_target_twilio,
            status=Notification.SCHEDULED,
            title='Test SMS',
            body="Whoops",
            scheduled_delivery=timezone.now() - timedelta(1),
        )

        self.notification_not_to_push = Notification.objects.create(
            target_user_record=self.user_target,
            status=Notification.DELIVERED,
            title="Title2",
            body="Body of the message2",
            scheduled_delivery=timezone.now() - timedelta(2),
            attempted_delivery=timezone.now() - timedelta(1),
        )

        self.email_notification_not_to_push = Notification.objects.create(
            target_user_record=self.user_target_email,
            status=Notification.DELIVERED,
            title="Title2",
            body="<h1>Body of the message</h1>",
            scheduled_delivery=timezone.now() - timedelta(2),
            attempted_delivery=timezone.now() - timedelta(1),
        )

        self.notification_with_invalid_extra = Notification.objects.create(
            target_user_record=self.user_target,
            status=Notification.SCHEDULED,
            title="Title3",
            body="Body of the message3",
            extra={"sound": "BAD_SOUND!"},
            scheduled_delivery=timezone.now() - timedelta(1),
        )

    def test_command__successful_push(self):
        """
        Verify a Notification that is SCHEDULED, has a scheduled_delivery that is earlier than
        the current datetime, and has valid 'extra' data gets pushed and updated correctly.
        """
        notification_to_push = Notification.objects.get(id=self.notification.id)
        self.assertEqual(notification_to_push.status, Notification.SCHEDULED)
        self.assertIsNone(notification_to_push.attempted_delivery)

        email_notification_to_push = Notification.objects.get(
            id=self.notification.id
        )
        self.assertEqual(email_notification_to_push.status, Notification.SCHEDULED)
        self.assertIsNone(email_notification_to_push.attempted_delivery)

        out = StringIO()
        call_command("process_notifications", stdout=out)

        notification_to_push = Notification.objects.get(id=self.notification.id)
        self.assertEqual(notification_to_push.status, Notification.DELIVERED)
        self.assertIsNotNone(notification_to_push.attempted_delivery)

        email_notification_to_push = Notification.objects.get(
            id=self.notification.id
        )
        self.assertEqual(email_notification_to_push.status, Notification.DELIVERED)
        self.assertIsNotNone(email_notification_to_push.attempted_delivery)

    def test_command__no_push(self):
        """
        Verify a Notification that is already DELIVERED doesn't get pushed or updated
        """
        notification_not_to_push = Notification.objects.get(
            id=self.notification_not_to_push.id
        )
        notification_not_to_push_attempted_delivery = notification_not_to_push.attempted_delivery
        self.assertEqual(notification_not_to_push.status, Notification.DELIVERED)
        self.assertEqual(
            notification_not_to_push.attempted_delivery, notification_not_to_push_attempted_delivery
        )

        email_notification_not_to_push = Notification.objects.get(
            id=self.email_notification_not_to_push.id
        )
        email_notification_not_to_push_attempted_delivery = email_notification_not_to_push.attempted_delivery
        self.assertEqual(
            email_notification_not_to_push.status, Notification.DELIVERED
        )
        self.assertEqual(
            email_notification_not_to_push.attempted_delivery, email_notification_not_to_push_attempted_delivery
        )

        out = StringIO()
        call_command("process_notifications", stdout=out)

        notification_not_to_push = Notification.objects.get(
            id=self.notification_not_to_push.id
        )
        self.assertEqual(notification_not_to_push.status, Notification.DELIVERED)
        self.assertEqual(
            notification_not_to_push.attempted_delivery, notification_not_to_push_attempted_delivery
        )

        email_notification_not_to_push = Notification.objects.get(
            id=self.email_notification_not_to_push.id
        )
        self.assertEqual(
            email_notification_not_to_push.status, Notification.DELIVERED
        )
        self.assertEqual(
            email_notification_not_to_push.attempted_delivery, email_notification_not_to_push_attempted_delivery
        )

    def test_command__failed_push(self):
        """
        Verify a Notification that is SCHEDULED, has a scheduled_delivery that is earlier than
        the current datetime, but has invalid 'extra' data doesn't get pushed and gets updated
        correctly.
        """
        notification_with_invalid_extra = Notification.objects.get(
            id=self.notification_with_invalid_extra.id
        )
        self.assertEqual(
            notification_with_invalid_extra.status, Notification.SCHEDULED
        )
        self.assertIsNone(notification_with_invalid_extra.attempted_delivery)

        out = StringIO()
        call_command("process_notifications", stdout=out)

        notification_with_invalid_extra = Notification.objects.get(
            id=self.notification_with_invalid_extra.id
        )
        self.assertEqual(notification_with_invalid_extra.status, Notification.RETRY)
        self.assertIsNotNone(notification_with_invalid_extra.attempted_delivery)

    def test_command__invalid_token(self):
        """
        Verify a Notification that is SCHEDULED, has a scheduled_delivery that is earlier than
        the current datetime, and has valid 'extra' data, but has an invalid token
        doesn't get pushed and updated correctly.
        """
        self.user_with_notifications.notification_target_user_records.all().all().delete()

        user_target_invalid_token = TargetUserRecord.objects.create(
            user=self.user_with_notifications,
            target=NotificationTarget.objects.get(name='Expo'),
            target_user_id="INVALID_TOKEN!!!!",
            description="Iphone 0",
        )

        notification_with_invalid_token = Notification.objects.create(
            target_user_record=user_target_invalid_token,
            status=Notification.SCHEDULED,
            title="Title4",
            body="Body of the message4",
            extra={
                "sound": "default",
                "data": {"Junk": "Junk"},
                "priority": "high",
                "ttl": "3",
                "expiration": "1486080000",
                "badge": "3",
            },
            scheduled_delivery="2018-10-24T12:42:13-04:00",
        )

        self.assertEqual(
            notification_with_invalid_token.status, Notification.SCHEDULED
        )
        self.assertIsNone(notification_with_invalid_token.attempted_delivery)

        out = StringIO()
        call_command("process_notifications", stdout=out)

        notification_with_invalid_token = Notification.objects.get(
            id=notification_with_invalid_token.id
        )
        self.assertEqual(notification_with_invalid_token.status, Notification.RETRY)
        self.assertIsNotNone(notification_with_invalid_token.attempted_delivery)

    def test_opt_out(self):
        user = User.objects.create_user(
            username="charlie",
            email="cbuckets@gmail.com",
            first_name="Charlie",
            last_name="Buckets",
            password="ImpressivePassword"
        )
        target_user_record = TargetUserRecord.objects.create(
            user=user,
            target=NotificationTarget.objects.get(name='Email'),
            target_user_id=user.email,
            description="Charlie email",
            active=True
        )
        Notification.objects.create(
            target_user_record=target_user_record,
            status=Notification.SCHEDULED,
            title="Notification to be recieved",
            body="<p>Body of the message</p>",
            scheduled_delivery="2018-10-24T12:42:13-04:00",
        )
        out = StringIO()
        call_command("process_notifications", stdout=out)

        Notification.objects.create(
            target_user_record=target_user_record,
            status=Notification.SCHEDULED,
            title="Notification to Not be recieved 1",
            body="<p>Body of the message</p>",
            scheduled_delivery="2018-10-24T12:42:13-04:00",
        )

        Notification.objects.create(
            target_user_record=target_user_record,
            status=Notification.SCHEDULED,
            title="Notification to Not be recieved 2",
            body="<p>Body of the message</p>",
            scheduled_delivery="2018-10-24T12:42:13-04:00",
        )

        NotificationOptOut.objects.create(user=user, active=True)

        call_command("process_notifications", stdout=out)

        notif_not_sent_1 = Notification.objects.get(
            target_user_record=target_user_record,
            title="Notification to Not be recieved 1")
        notif_not_sent_2 = Notification.objects.get(
            target_user_record=target_user_record,
            title="Notification to Not be recieved 2")

        self.assertEqual(notif_not_sent_1.status, "OPTED OUT")
        self.assertEqual(notif_not_sent_2.status, "OPTED OUT")

    def test_command__inactive_target(self):
        """
        Verify a Notification scheduled to be sent to an inactive device is marked as INACTIVE_DEVICE and ignored
        """
        notification_to_push = Notification.objects.get(id=self.notification.id)
        target_user_record = TargetUserRecord.objects.get(
            user=notification_to_push.target_user_record.user, target__name='Expo')
        target_user_record.active = False
        target_user_record.save()
        notification_to_push.user_target = target_user_record
        notification_to_push.save()

        self.assertEqual(notification_to_push.status, Notification.SCHEDULED)
        self.assertIsNone(notification_to_push.attempted_delivery)

        out = StringIO()
        call_command("process_notifications", stdout=out)

        notification_to_push = Notification.objects.get(id=self.notification.id)
        self.assertEqual(notification_to_push.status, Notification.INACTIVE_DEVICE)
        self.assertIsNone(notification_to_push.attempted_delivery)

        self.assertEqual(
            Notification.objects.filter(
                target_user_record=notification_to_push.target_user_record,
                title=notification_to_push.title,
                body=notification_to_push.body,
                extra=notification_to_push.extra,
                scheduled_delivery=notification_to_push.scheduled_delivery
            ).count(),
            1,
        )

    def test_deactivates_on_device_not_registered(self):
        """
        Test that a user target is deactivated
        when Expo/Firebase Cloud Messaging returns
        a device not registered error.
        Returns
        -------

        """
        self.assertEqual(self.dev_user_target.active, True)
        response = PushResponse(
            push_message="",
            status=PushResponse.ERROR_STATUS,
            message='"adsf" is not a registered push notification recipient',
            details={'error': PushResponse.ERROR_DEVICE_NOT_REGISTERED}
        )
        handle_push_response(self.dev_notification, response=response)
        self.dev_user_target.refresh_from_db()
        self.assertEqual(self.dev_user_target.active, False)



"""Admin interface setup for Notifications feature/app"""

from django.contrib import admin
from django.utils.safestring import mark_safe

from .models import (
    Notification,
    NotificationOptOut,
    NotificationTarget,
    TargetUserRecord,
)
from .utils.admin_site_utils import (
    MeOrAllFilter,
    is_null_filter_factory,
    USER_SEARCH_FIELDS,
)


@admin.register(Notification)
class NotificationAdmin(admin.ModelAdmin):
    list_display = [
        "target_user_record",
        "status",
        "scheduled_delivery",
        "attempted_delivery",
    ]
    list_filter = [
        "status",
        is_null_filter_factory("attempted_delivery"),
        "target_user_record__target",
    ]

    search_fields = [
        "target_user_record__" + field for field in USER_SEARCH_FIELDS
    ] + ["title", "body"]

    autocomplete_fields = ["target_user_record"]


@admin.register(NotificationOptOut)
class NotificationOptOutAdmin(admin.ModelAdmin):
    list_display = ["user", "active"]
    autocomplete_fields = ["user"]
    search_fields = USER_SEARCH_FIELDS


@admin.register(NotificationTarget)
class NotificationTargetAdmin(admin.ModelAdmin):
    list_display = ["name", "notification_module_name"]


@admin.register(TargetUserRecord)
class UserInNotificationTargetAdmin(admin.ModelAdmin):
    list_display = [
        "user",
        "target",
        "description",
        "target_user_id",
        "active",
    ]

    list_filter = ["active", MeOrAllFilter, "target"]

    search_fields = USER_SEARCH_FIELDS + (
        "target_user_id",
        "description",
    )



import uuid

from django.conf import settings
from django.db import models

from .abstract import CreatedModifiedAbstractModel
from .notification import Notification


class NotificationOptOut(CreatedModifiedAbstractModel):
    """
    Definition of a User Opt-Out Model.

    Users who have opted-out of communications will have an instance of this model.

    Attributes
    ----------
    id : UUID
        The unique UUID of the record.
    user : Django User / Custom User Instance
        The User/Custom User instance associated with this record.
    active : boolean
        Indicator for whether the opt out is active or not.
    """

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    user = models.OneToOneField(
        settings.AUTH_USER_MODEL,
        on_delete=models.CASCADE,
        related_name="notification_opt_out",
    )
    active = models.BooleanField(default=False)

    class Meta:
        db_table = "notification_system_opt_out"
        verbose_name_plural = "Notification Opt Outs"

    def __str__(self):
        return self.user.username

    def save(self, *args, **kwargs):
        """
        When an instance of this model is saved, if the opt out is active
        change the status of notifications with a current status of
        SCHEDULED or RETRY to OPTED_OUT.
        """
        if self.active:
            Notification.objects.filter(
                status__in=[Notification.SCHEDULED, Notification.RETRY],
                target_user_record__user=self.user,
            ).update(status=Notification.OPTED_OUT)
        super(NotificationOptOut, self).save(*args, **kwargs)



"""Custom exceptions for Django Notification System App."""


class NotificationSystemError(Exception):
    """Parent class for custom exceptions."""

    pass


class NotificationsNotCreated(NotificationSystemError):
    """Exception to raise when one or more notifications cannot be created."""

    pass


class NotificationNotSent(NotificationSystemError):
    """
    Exception to raise when a notification is not able
    to be sent, but we do not have a more specific cause.
    """

    pass


class UserIsOptedOut(NotificationNotSent):
    """
    Exception to raise when a notification is not able to
    be sent because the user is opted out.
    """

    def __init__(self):
        super().__init__("User is opted out")


class UserHasNoTargetRecords(NotificationNotSent):
    """
    Exception to raise when a notification is not able to
    be sent because the user has no available targets.
    """

    def __init__(self):
        super().__init__("User has no active targets")


"""Django Management Command."""
from django.contrib.auth import get_user_model
from django.core.management.base import BaseCommand

from ...models import TargetUserRecord, NotificationTarget

User = get_user_model()


class Command(BaseCommand):
    """
    Create Email UserTargets for all users in the DB.
    """

    help = __doc__

    def handle(self, *args, **options):
        """
        This is what is being run by manage.py
        """
        all_users = User.objects.all()

        email_target = NotificationTarget.objects.get(
            name='Email', notification_module_name='email')

        for user in all_users:
            # Calling this will create an email user target for each user.
            if user.email:
                TargetUserRecord.objects.update_or_create(
                    user=user,
                    target=email_target,
                    target_user_id=user.email,
                    defaults={
                        "active": True,
                        "description": f"{user.first_name} {user.last_name}'s Email",
                    },
                )
            else:
                print(f"{user.username} has no email address on record.")


from collections import namedtuple
from exponent_server_sdk import PushResponseError, DeviceNotRegisteredError, MessageTooBigError, MessageRateExceededError, PushServerError, PushMessage, PushResponse


class MockPushClient(object):
    """Exponent push client

    See full API docs at https://docs.expo.io/versions/latest/guides/push-notifications.html#http2-api
    """
    DEFAULT_HOST = "https://exp.host" # TODO
    DEFAULT_BASE_API_URL = "/--/api/v2"

    def __init__(self, host=None, api_url=None):
        """Construct a new PushClient object.

        Args:
            host: The server protocol, hostname, and port.
            api_url: The api url at the host.
        """
        self.host = host
        if not self.host:
            self.host = MockPushClient.DEFAULT_HOST

        self.api_url = api_url
        if not self.api_url:
            self.api_url = MockPushClient.DEFAULT_BASE_API_URL

    @classmethod
    def is_exponent_push_token(cls, token):
        """Returns `True` if the token is an Exponent push token"""
        import six

        return (
            isinstance(token, six.string_types) and
            token.startswith('ExponentPushToken'))

    def _publish_internal(self, push_messages):
        """Send push notifications

        The server will validate any type of syntax errors and the client will
        raise the proper exceptions for the user to handle.

        Each notification is of the form:
        {
          'to': 'ExponentPushToken[xxx]',
          'body': 'This text gets display in the notification',
          'badge': 1,
          'data': {'any': 'json object'},
        }

        Args:
            push_messages: An array of PushMessage objects.
        """
        receipts = []
        for i, message in enumerate(push_messages):
            payload = message.get_payload()
            receipts.append(PushResponse(
                push_message=message,
                status=PushResponse.SUCCESS_STATUS,
                message='',
                details=None))
            if payload.get('sound', 'default') != 'default':
                raise PushServerError('Request failed', {})

        return receipts

    def publish(self, push_message):
        """Sends a single push notification

        Args:
            push_message: A single PushMessage object.

        Returns:
           A PushResponse object which contains the results.
        """
        return self.publish_multiple([push_message])[0]

    def publish_multiple(self, push_messages):
        """Sends multiple push notifications at once

        Args:
            push_messages: An array of PushMessage objects.

        Returns:
           An array of PushResponse objects which contains the results.
        """
        return self._publish_internal(push_messages)



from setuptools import setup, find_packages

def read(f):
    return open(f, encoding='utf-8').read()

setup(
    # This is the name of your project. The first time you publish this
    # package, this name will be registered for you. It will determine how
    # users can install this project, e.g.:
    #
    # $ pip install sampleproject
    #
    # And where it will live on PyPI: https://pypi.org/project/sampleproject/
    #
    # There are some restrictions on what makes a valid project name
    # specification here:
    # https://packaging.python.org/specifications/core-metadata/#name
    name='django-notification-system',  # Required

    # Versions should comply with PEP 440:
    # https://www.python.org/dev/peps/pep-0440/
    #
    # For a discussion on single-sourcing the version across setup.py and the
    # project code, see
    # https://packaging.python.org/en/latest/single_source_version.html
    version='1.5.1',  # Required

    # This is a one-line description or tagline of what your project does. This
    # corresponds to the "Summary" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#summary
    description='Notification functionality to use within Django',  # Required

    long_description=read('README.md'),
    long_description_content_type='text/markdown',

    # This should be a valid link to your project's main homepage.
    #
    # This field corresponds to the "Home-Page" metadata field:
    # https://packaging.python.org/specifications/core-metadata/#home-page-optional
    url='https://github.com/crcresearch/django-notification-system',  # Optional

    # You can just specify package directories manually here if your project is
    # simple. Or you can use find_packages().
    #
    # Alternatively, if you just want to distribute a single Python file, use
    # the `py_modules` argument instead as follows, which will expect a file
    # called `my_module.py` to exist:
    #
    #   py_modules=["my_module"],
    #
    packages=find_packages(include=['django_notification_system', 'django_notification_system.*']),  # Required

    # This field lists other packages that your project depends on to run.
    # Any package you put here will be installed by pip when your project is
    # installed, so they must be valid existing projects.
    #
    # For an analysis of "install_requires" vs pip's requirements files see:
    # https://packaging.python.org/en/latest/requirements.html
    install_requires=[
        'Django>=3.1.3',
        'html2text==2018.1.9',
        'twilio==6.29.1',
        'exponent_server_sdk==0.3.1']
)


from django.contrib.auth.models import User
from django.core.management import call_command
from django.test import TestCase
from six import StringIO

from django_notification_system.management.commands.create_email_target_user_records import Command
from django_notification_system.models import TargetUserRecord

cmd = Command()


class TestCommand(TestCase):
    def setUp(self):
        self.user = User.objects.create_user(
            username="Danglesauce",
            email="danglesauce@gmail.com",
            first_name="Dangle",
            last_name="Sauce",
            password="ImpressivePassword")

    def test_create_email_target_user_records(self):
        """
        Ensure command is functioning correctly.
        """
        # Delete UserInNotificationTarget from initial signal.
        TargetUserRecord.objects.all().delete()

        pre_call = TargetUserRecord.objects.all()
        self.assertEqual(len(pre_call), 0)

        out = StringIO()
        call_command('create_email_target_user_records', stdout=out)

        post_call = TargetUserRecord.objects.all()
        self.assertEqual(len(post_call), 1)

        for target in post_call:
            if target.user == self.user:
                self.assertEqual(target.target_user_id,
                                 self.user.email)




from django.contrib.auth.models import User
from django.test.testcases import TestCase


from django_notification_system.models import (
    Notification, NotificationTarget, TargetUserRecord)
from django_notification_system.notification_creators.email import (
    create_notification)


class TestCreateEmailNotification(TestCase):
    def setUp(self):
        self.user_with_targets = User.objects.create_user(
            username='sadboi@gmail.com',
            email='sadboi@gmail.com',
            first_name='Sad',
            last_name='Boi',
            password='Ok.')

        self.user_without_target = User.objects.create_user(
            username='skeeter@gmail.com',
            email='skeeter@gmail.com',
            first_name='Skeeter',
            last_name='Skeetington',
            password='Sure.')

        # We call this to create the email user_target.
        self.user_with_targets.save()

        self.target, created = NotificationTarget.objects.get_or_create(
            name='Email',
            notification_module_name='email'
        )

        self.target_user_record = TargetUserRecord.objects.create(
            user=self.user_with_targets,
            target=self.target,
            target_user_id='sadboi@gmail.com',
            description="Sad Boi's Email",
            active=True
        )

    def test_successfully_create_notifications(self):
        """
        This test checks that email notifications are created for all
        email user targets associated with the user.
        """
        pre_function_notifications = Notification.objects.all()
        self.assertEqual(len(pre_function_notifications), 0)

        create_notification(
            user=self.user_with_targets,
            title="Hi.",
            body="Hello there, friend.")

        post_function_notifications = Notification.objects.all()
        self.assertEqual(len(post_function_notifications), 1)



from dateutil.relativedelta import relativedelta

from django.contrib.auth.models import User
from django.utils import timezone

from django_notification_system.exceptions import UserIsOptedOut
from django_notification_system.models.target_user_record import (
    TargetUserRecord,
)


def check_for_user_opt_out(user: User):
    """Determine if a user has an active opt-out.

    Args:
        user (User): The user to perform the check on.

    Raises:
        UserIsOptedOut: If the user has an active opt out.
    """
    if (
        hasattr(user, "notification_opt_out")
        and user.notification_opt_out.active
    ):
        raise UserIsOptedOut


def user_notification_targets(user: User, target_name: str):
    """Return all active user notifications targets for a given notification target.

    Args:
        user (User): The user to retrieve user targets for.
        target_name (str): The name of the target to retrieve user targets for.

    Returns:
        [UserInNotificationTarget]: A Django queryset of UserInNotificationTarget instances.
    """
    return TargetUserRecord.objects.filter(
        user=user,
        target__name=target_name,
        active=True,
    )


def check_and_update_retry_attempts(notification, minute_interval=None):
    """
    Check if the retry_attempt and max_retries are equal.
    If they are we want to change the status to DELIVERY_FAILURE.
    If not, we want to RETRY at a given minute interval.
    """
    if notification.max_retries != notification.retry_attempts:
        notification.retry_attempts += 1
        notification.save()
        # We have to go deeper...
        if notification.max_retries != notification.retry_attempts:
            notification.status = notification.RETRY
            if minute_interval is not None:
                notification.scheduled_delivery = (
                    timezone.now() + relativedelta(minutes=minute_interval)
                )
            else:
                notification.scheduled_delivery = (
                    timezone.now()
                    + relativedelta(minutes=notification.retry_time_interval)
                )
            notification.attempted_delivery = timezone.now()
            notification.save()

        else:
            notification.status = notification.DELIVERY_FAILURE
            notification.attempted_delivery = timezone.now()
            notification.save()



import uuid

from django.core.exceptions import ValidationError
from django.db import models

from .abstract import CreatedModifiedAbstractModel
from .target_user_record import TargetUserRecord


class Notification(CreatedModifiedAbstractModel):
    """
    Definition of a Notification.

    Attributes
    ----------
    id : UUID
        The unique UUID of the record.
    target_user_record : UserInNotificationTarget
        The UserInNotificationTarget associated with notification
    title : str
        The title for the notification. Exact representation depends on the target.
        For example, for an email notification this will be used as the subject of the email.
    body : str
        The main message of the notification to be sent.
    extra : dict
        A dictionary of extra data to be sent to the notification processor. Valid keys
        are determined by each processor.
    status : CharField
        The status of Notification. Options are: 'SCHEDULED', 'DELIVERED', 'DELIVERY_FAILURE', 'RETRY', 'INACTIVE_DEVICE'
    scheduled_delivery : DateTimeField
        Day and time Notification is to be sent.
    attempted_delivery : DateTImeField
        Day and time attempted to deliver Notification.
    retry_time_interval : PositiveIntegerField
        If a notification fails, this is the amount of time to wait until retrying to send it.
    retry_attempts : PositiveIntegerField
        The number of retries that have been attempted.
    max_retries : PositiveIntegerField
        The max number of allowed retries.
    """

    DELIVERED = "DELIVERED"
    DELIVERY_FAILURE = "DELIVERY FAILURE"
    INACTIVE_DEVICE = "INACTIVE DEVICE"
    OPTED_OUT = "OPTED OUT"
    RETRY = "RETRY"
    SCHEDULED = "SCHEDULED"

    STATUS_CHOICES = (
        (DELIVERED, "Delivered"),
        (DELIVERY_FAILURE, "Delivery Failure"),
        (INACTIVE_DEVICE, "Inactive Device"),
        (OPTED_OUT, "Opted Out"),
        (RETRY, "Retry"),
        (SCHEDULED, "Scheduled"),
    )

    id = models.UUIDField(primary_key=True, default=uuid.uuid4, editable=False)
    target_user_record = models.ForeignKey(
        TargetUserRecord,
        on_delete=models.CASCADE,
        related_name="notifications",
    )
    title = models.CharField(max_length=100)
    body = models.TextField()
    extra = models.JSONField(blank=True, null=True, default=dict)
    status = models.CharField(max_length=16, choices=STATUS_CHOICES)
    scheduled_delivery = models.DateTimeField()
    attempted_delivery = models.DateTimeField(null=True, blank=True)
    retry_time_interval = models.PositiveIntegerField(default=0)
    retry_attempts = models.PositiveIntegerField(default=0)
    max_retries = models.PositiveIntegerField(default=3)

    class Meta:
        db_table = "notification_system_notification"
        verbose_name_plural = "Notifications"
        unique_together = [
            "target_user_record",
            "scheduled_delivery",
            "title",
            "extra",
        ]

    def __str__(self):
        return "{} - {} - {}".format(
            self.target_user_record.user.username,
            self.status,
            self.scheduled_delivery,
        )

    def clean(self):
        """
        Perform a few data checks whenever an instance is saved.

        1. Don't allow notifications with an attempted delivery date to
           have a status of 'SCHEDULED'.
        2. If a notification has a status other than 'SCHEDULED' it MUST
           have an attempted delivery date.
        3. Don't allow notifications to be saved if the user has opted out.

        Raises
        ------
        ValidationError
            Will include details of what caused the validation error.
        """
        opted_out = (
            hasattr(self.target_user_record.user, "notification_opt_out")
            and self.target_user_record.user.notification_opt_out.has_opted_out
        )
        if opted_out:
            raise ValidationError("This user has opted out of Notifications.")

        if self.attempted_delivery and self.status == "SCHEDULED":
            raise ValidationError(
                "Status cannot be 'SCHEDULED' if there is an attempted delivery."
            )

        if not self.attempted_delivery and self.status not in ["SCHEDULED", "OPTED OUT"]:
            raise ValidationError(
                "Attempted Delivery must be filled out if Status is {}".format(
                    self.status
                )
            )
